/* Triangular number solution to FIFO State Pen problem.

   Part of example code for "Drill Bits" column in ACM Queue magazine
   November/December 2023,  https://queue.acm.org/DrillBits

   Copyright (C) 2023  Terence Kelly
   Contact:  tpkelly @ { acm.org, cs.princeton.edu, eecs.umich.edu }

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.

   See the README file for notes on the design rationale.
*/

#ifdef __APPLE__
#error "MacOS reportedly lacks a required POSIX 2008 feature; see README"
#endif

#define _POSIX_C_SOURCE 200809L  // for scanf("%ms")

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"

int main(int argc, char *argv[]) {
  char *name, *escapee, **I;  // I[] maps sernos->names; I[0] unused
  uint64_t T0, T1, sum = 0, TA, i, esn, A = 0;  // number of arrivals
  uintmax_t capacity;
  int flag = 0;

  if (3 != argc)
    BAIL("usage:  %s  capacity  escapee\n", argv[0]);
  capacity = s2u(argv[1]);
  escapee  =     argv[2];
  if (0 >= capacity)
    BAIL("bad capacity %" PRIuMAX "\n", capacity);

  if (NULL == (I = (char **)calloc(1 + capacity, sizeof *I)))
    BAIL("allocation failure:  %s\n", strerror(errno));

  errno = 0;
  T0 = Tns();
  // scanf() 'm' is POSIX 2008 but reportedly unsupported on MacOS
  while (1 == scanf("%ms\n", &name)) {
    if (capacity < ++A)
      BAIL("inputs exceed capacity %" PRIuMAX "\n", capacity);
    if (strcmp(name, escapee))
      I[A] = name;
    else {
      if (flag++)
        BAIL("escapee name '%s' on input twice\n", escapee);
      free(name);
      I[A] = escapee;  // for fast pointer comparisons later
    }
  }
  T1 = Tns();
  if (A != capacity)
    PE("WARNING:  A == %" PRIu64 " != capacity\n", A);
  if (errno)  // scanf() 'm' allocation might fail
    BAIL("errno => '%s'\n", strerror(errno));
  if (! flag)
    BAIL("escapee name '%s' not on input\n", escapee);
  P("time to read input: %" PRIu64 "\n", T1 - T0);

  T0 = Tns();
  if (UINT32_MAX <= A)
    BAIL("overflow would occur because A == %" PRIu64 "\n", A);
  TA = A * (A+1) / 2;  // triangular number
  for (i = 1; i <= A; i++)
    if (I[i] != escapee)  // pointer comparison, not strcmp
      sum += i;
  esn = TA - sum;
  T1 = Tns();
  P("TIME triangular %" PRIuMAX " %" PRIu64 "\n", capacity, T1 - T0);
  P("escapee serno %" PRIu64 " name '%s'\n", esn, I[esn]);
  assert(I[esn] == escapee);

  return 0;
}

