/* Utility functions used by several programs (implementation).

   Part of example code for "Drill Bits" column in ACM Queue magazine
   November/December 2023,  https://queue.acm.org/DrillBits

   Copyright (C) 2023  Terence Kelly
   Contact:  tpkelly @ { acm.org, cs.princeton.edu, eecs.umich.edu }

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#define _POSIX_C_SOURCE 199309  // for clock_gettime()

#include <ctype.h>
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "util.h"

// time in nanoseconds
uint64_t Tns(void) {
  struct timespec ts;
  if (clock_gettime(CLOCK_MONOTONIC, &ts))   // _RAW ?
    BAIL("clock_gettime failed: %s\n", strerror(errno));
  return   (uint64_t)ts.tv_sec * UINT64_C(1000000000)
         + (uint64_t)ts.tv_nsec;
}

// careful string -> int conversion
uintmax_t s2u(const char *s) {
  char *ep;
  uintmax_t r;
  if (NULL == s || '\0' == *s)
    BAIL("NULL or zero-length string\n");
  // isdigit() check below is necessary because strtoumax() on Linux
  // incorrectly handles negative inputs ("-3").  Instead of checking
  // every possible troublesome special case, we insist that every
  // input char must be a numeric digit.  This bans explicitly
  // positive inputs ("+3").
  for (const char *p = s; *p; p++)
    if (0 > *p || ! isdigit((unsigned char)*p))
      BAIL("bad char '%c' at position %td in '%s'\n", *p, p-s, s);
  errno = 0;
  r = strtoumax(s, &ep, 10);
  if ('\0' != *ep)
    BAIL("conversion failure '%s'\n", s);
  if (UINTMAX_MAX == r && errno)
    BAIL("out of range or invalid '%s'\n", s);
  return r;
}

