/* Bitmask based solution to FIFO State Pen problem.

   Copyright (C) 2023  John Dilley <jad@ieee.org>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.

   See the README file for notes on the design rationale.
*/

#define _POSIX_C_SOURCE 200809L  // necessary on Ubuntu for strdup()

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "util.h"

////
// numBits(val) -- return the number of one bits in val (Brian Kernighan's algorithm)
int numBits(uint64_t val) {
  int n = 0;
  while (val) {
    val &= val-1;
    n++;
  }
  return n;
}

int main(int argc, char *argv[]) {
  char name[32], *escapee, **I;			// names, indexed by serno, zero-based
  uint64_t T0, T1;				// for timing 
  uint64_t A = 0;				// A: number of arrivals
  int64_t esn = -1;				// escapee number
  uint64_t *B;					// 64 bit cell blocks
  uintmax_t capacity;				// max population
  uint64_t nBlocks;				// number of cell blocks holding that capacity
  int flag = 0;

  if (3 != argc)
    BAIL("usage:  %s  capacity  escapee\n", argv[0]);
  capacity = s2u(argv[1]);
  escapee  =     argv[2];

  if (0 >= capacity)
    BAIL("bad capacity %" PRIuMAX "\n", capacity);

  if (NULL == (I = (char **)calloc(1 + capacity, sizeof *I)))
    BAIL("allocation failure I:  %s\n", strerror(errno));
  nBlocks  = 1 + (capacity >> 6);
  if (NULL == (B = (uint64_t *)calloc(nBlocks, sizeof *B)))
    BAIL("allocation failure B:  %s\n", strerror(errno));

  errno = 0;
  T0 = Tns();
  uint64_t cell = 0, block;
  while (1 == scanf("%s\n", name)) {
    if (capacity < A)
      BAIL("too many inputs, capacity %" PRIuMAX "\n", capacity);

    cell = A & 0x3F;				// cell is last 6 bits
    block = A >> 6;				// blocks of 64

    I[A] = strdup(name);
    if (NULL == I[A])
      BAIL("strdup failed, errno => '%s'\n", strerror(errno));
    if (! strcmp(name, escapee)) {
      if (flag++)
	BAIL("name '%s' on input twice\n", escapee);
    } else {
      B[block] |= (uint64_t)1<<cell;
    }
    A++;
  }
  // set the rest of the cells in the last cell block to 1 (no escapee)
  ++cell;
  while (cell < 64)
    B[A>>6] |= (uint64_t)1<<(cell++);

  T1 = Tns();
  if (errno)  // scanf("%ms") allocation might fail. Or strdup or ...
    BAIL("errno => '%s'\n", strerror(errno));
  if (! flag)
    BAIL("name '%s' not on input\n", escapee);
  if (A != capacity)
    PE("WARNING:  A == %" PRIu64 " != capacity\n", A);
  P("time to read input: %" PRIu64 "\n", T1 - T0);

  T0 = Tns();
  uint64_t fullblock = UINT64_MAX;
  for (uint64_t b = 0; b < nBlocks; b++) {
    if (B[b] != fullblock) {
      if (numBits(~B[b]) != 1)
	BAIL("more than one escapee in this cell block!\n");
      esn = (b<<6) + (int)log2(~B[b]);
      break;
    }
  }
  if (-1 == esn)
    BAIL("esn not found\n");
  
  T1 = Tns();
  P("TIME cellblocks %" PRIuMAX " %" PRIu64 "\n", capacity, T1 - T0);
  P("escapee serno %" PRIu64 ", name '%s'\n", esn+1, I[esn]);
  assert(strcmp(I[esn], escapee) == 0);

  // quiet valgrind (Linux) or leaks (Mac) by freeing malloc'ed memory
  T0 = Tns();
  free(B);
  for (uint64_t n = 0; n < A; n++)
    free(I[n]);
  free(I);
  T1 = Tns();
  P("Cleanup took %" PRIu64 " nsec \n", T1 - T0);
  
  return 0;
}

