/* Performance tests for certain key operations

   Copyright (C) 2023  John Dilley <jad@ieee.org>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.

   See the README file for notes on the design rationale.
*/

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "util.h"

////
// Test alternative triangular number calculations: the A*(A+1)/2 canonical, and an
// equivalent version less susceptible to integer overflow.
double testTaCalc(uint64_t n)
{
  uint64_t TA, TS;
  uint64_t T0, T1;				// nanosec timer results
  int64_t Test1, Test2;				// timer results from the loops
  
  TS = 0;
  T0 = Tns();
  for (uint64_t i = 0; i < n; i++) {
    TA = i * (i+1) / 2;				// compute the triangular number
    TS += TA;					// don't optimize away the loop
  }
  T1 = Tns();
  Test1 = T1 - T0;
  P("Test1 took %7.03f msec, %.02lf%% full (check for overflow)\n", Test1/1e6, ((double)TS/(double)UINT64_MAX)*100);

  TS = 0;
  T0 = Tns();
  for (uint64_t i = 0; i < n; i++) {
    TA = (i+1)/2 * i + ((i+1)&01 ? i/2 : 0);	// reduced overflow calculation
    TS += TA;					// don't optimize away
  }
  T1 = Tns();
  Test2 = T1 - T0;
  P("Test2 took %7.03f msec, %.02lf%% full (check for overflow)\n", Test2/1e6, ((double)TS/(double)UINT64_MAX)*100);

  double diffMsec = (Test2-Test1)/1e6;
  double pctDiff  = 1e2*(Test2-Test1)/Test1;
  P("Test2 was  %7.03f msec (%.02f%%) %s\n", fabs(diffMsec), fabs(pctDiff), pctDiff>0?"slower":"FASTER");

  return diffMsec;
}

////
// Basic version of a log2 calculation that returns the index of the highest bit set.
// It's an accurate log2 value for values with just one bit set; low for all others.
inline double myLog2(uint64_t val)
{
  uint64_t n = 0;
  while (val >>= 1)
    n++;
  return n;
}

////
// DeBruijn-like function to compute log base 2 of a 64 bit value with only one bit set.
// For numbers with more than one bit set it's approximate (lower than the true log2).
// It's much faster than the builtin log2 and much much faster than the basic approach.
int log2_64 (uint64_t value)
{
  const int logTable64[64] = {
    63,  0, 58,  1, 59, 47, 53,  2,
    60, 39, 48, 27, 54, 33, 42,  3,
    61, 51, 37, 40, 49, 18, 28, 20,
    55, 30, 34, 11, 43, 14, 22,  4,
    62, 57, 46, 52, 38, 26, 32, 41,
    50, 36, 17, 19, 29, 10, 13, 21,
    56, 45, 25, 31, 35, 16,  9, 12,
    44, 24, 15,  8, 23,  7,  6,  5};

    value |= value >> 1;
    value |= value >> 2;
    value |= value >> 4;
    value |= value >> 8;
    value |= value >> 16;
    value |= value >> 32;
    return logTable64[((uint64_t)((value - (value >> 1))*0x07EDD5E59A4E28C2)) >> 58];
    // Credit to https://stackoverflow.com/questions/11376288/fast-computing-of-log2-for-64-bit-integers
}


////
// Test alternative calculations for log to the base 2.  Return the difference between
// the third (DeBruijn) and first (builtin math.log2() from libc).
double testLogCalc(uint64_t n)
{
  uint64_t T0, T1;				// nanosec timer results
  int64_t Test1, Test2, Test3;			// timer results from the loops
  double accumulator;				// prevent optimizer skipping loop

  accumulator = 0;
  T0 = Tns();
  for (uint64_t i = 1; i <= n; i++) {		// can't take log of zero...
    double x = log2((double)i);
    accumulator += x;
  }
  T1 = Tns();
  Test1 = T1 - T0;
  P("Test1 took %7.03f msec, sum was %lf\n", Test1/1e6, accumulator/1e6);
  // I know the accumulator makes no sense; just shuts up a compiler warning

  accumulator = 0;
  T0 = Tns();
  for (uint64_t i = 1; i <= n; i++) {
    double x = myLog2(i);			// doesn't do full log calc, only first digit
    accumulator += x;
  }
  T1 = Tns();
  Test2 = T1 - T0;
  P("Test2 took %7.03f msec, sum was %lf\n", Test2/1e6, accumulator/1e6);
  double diffMsec = (Test2-Test1)/1e6;
  double pctDiff  = 1e2*(Test2-Test1)/Test1;
  P("Test2 was  %7.03f msec (%.02f%%) %s than log2(Test1)\n", fabs(diffMsec), fabs(pctDiff), pctDiff>0?"slower":"FASTER");

  accumulator = 0;
  T0 = Tns();
  for (uint64_t i = 1; i <= n; i++) {
    double x = log2_64(i);			// DeBruijn-like log approximation
    // NOTE: it doesn't do full log calc, only log2 up to the first digit
    accumulator += x;
  }
  T1 = Tns();
  Test3 = T1 - T0;
  P("Test3 took %7.03f msec, sum was %lf\n", Test3/1e6, accumulator/1e6);

  diffMsec = (Test3-Test1)/1e6;
  pctDiff  = 1e2*(Test2-Test1)/Test1;
  P("Test3 was  %7.03f msec (%.02f%%) %s than log2(Test1)\n", fabs(diffMsec), fabs(pctDiff), pctDiff>0?"slower":"FASTER");

  return diffMsec;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  Main routine
///////////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[]) {
  uint64_t numTests;
  
  if (2 != argc)
    BAIL("usage:  %s  numTests\n", argv[0]);
  numTests = s2u(argv[1]);

  ////
  // Test performance of two triangular TA calculations
  uint64_t size = 1 << 20;			// start at about a billion
  double diffMsec = 0;
  for (uint64_t n = 1; n <= numTests; n++) {
    P("\nTesting TA calc with %" PRIu64 " (0x%" PRIx64 ") trials, test %" PRIu64 " of %" PRIu64 "\n",
      size, size, n, numTests);
    diffMsec += testTaCalc(size);
    size += 1 << 20;				// increase about a billion each time
  }
  P("\nTA Test2 was %7.03f msec %s on average\n", fabs(diffMsec)/numTests, diffMsec>0?"slower":"FASTER");

  ////
  // Test performance of log2 versus integer only replacement
  size = 1 << 20;				// start at about a billion
  diffMsec = 0;
  for (uint64_t n = 1; n <= numTests; n++) {
    P("\nTesting log2 calc with %" PRIu64 " (0x%" PRIx64 ") trials, test %" PRIu64 " of %" PRIu64 "\n",
      size, size, n, numTests);
    diffMsec += testLogCalc(size);
    size += 1 << 20;				// increase about a billion each time
  }
  P("\nlog2_64(Test3) was %7.03f msec %s on average than builtin log2(Test1)\n",
    fabs(diffMsec)/numTests, diffMsec>0?"slower":"FASTER");
  
  return 0;
}
